/* Soot - a J*va Optimization Framework
 * Copyright (C) 2002 Ondrej Lhotak
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

package soot.jimple.spark.sets;
import soot.jimple.spark.*;
import soot.jimple.spark.pag.Node;
import soot.jimple.spark.pag.PAG;
import soot.jimple.spark.internal.*;
import java.util.*;
import soot.Type;

/** Implementation of points-to set using a bit vector.
 * @author Ondrej Lhotak
 */
public final class BitPointsToSet extends PointsToSetInternal {
    public BitPointsToSet( Type type, PAG pag ) {
        super( type );
        this.pag = pag;
        if( nodes == null ) {
            SIZE = pag.getNumAllocNodes()/64+2;
            nodes = new Node[ SIZE * 64 ];
        }
        bits = new long[ SIZE ];
    }
    /** Returns true if this set contains no run-time objects. */
    public final boolean isEmpty() {
        return empty;
    }
    /** Adds contents of other into this set, returns true if this set 
     * changed. */
    public final boolean addAll( final PointsToSetInternal other,
            final PointsToSetInternal exclude ) {
        boolean ret = false;
        long[] mask = null;
        TypeManager typeManager = pag.getTypeManager();
        if( !typeManager.castNeverFails( other.getType(), this.getType() ) ) {
            mask = typeManager.get( this.getType() );
        }
        if( other instanceof BitPointsToSet ) {
            BitPointsToSet o = (BitPointsToSet) other;
            if( exclude == null || exclude.isEmpty() ) {
                if( mask == null ) {
                    for( int i=0; i < SIZE; i++ ) {
                        long l = o.bits[i] & ~bits[i];
                        if( l != 0L ) {
                            ret = true;
                            empty = false;
                            bits[i] |= l;
                        }
                    }
                } else {
                    for( int i=0; i < SIZE; i++ ) {
                        long l = o.bits[i] & ~bits[i] & mask[i];
                        if( l != 0L ) {
                            ret = true;
                            empty = false;
                            bits[i] |= l;
                        }
                    }
                }
                return ret;
            } else if( exclude instanceof BitPointsToSet ) {
                BitPointsToSet e = (BitPointsToSet) exclude;
                if( mask == null ) {
                    for( int i=0; i < SIZE; i++ ) {
                        long l = o.bits[i] & ~bits[i] & ~e.bits[i];
                        if( l != 0L ) {
                            ret = true;
                            empty = false;
                            bits[i] |= l;
                        }
                    }
                } else {
                    for( int i=0; i < SIZE; i++ ) {
                        long l = o.bits[i] & ~bits[i] & ~e.bits[i] & mask[i];
                        if( l != 0L ) {
                            ret = true;
                            empty = false;
                            bits[i] |= l;
                        }
                    }
                }
                return ret;
            } else {
                if( mask == null ) {
                    for( int i=0; i < SIZE; i++ ) {
                        long l = o.bits[i] & ~bits[i];
                        if( l != 0L ) {
                            for( int j=0; j<64; j++ ) {
                                if( ( l & (1L<<j) ) != 0  ) {
                                    Node n = nodes[i*64+j];
                                    if( exclude.contains( n ) ) {
                                        l &= ~(1L<<j);
                                    }
                                }
                            }
                            if( l != 0L ) {
                                ret = true;
                                empty = false;
                                bits[i] |= l;
                            }
                        }
                    }
                } else {
                    for( int i=0; i < SIZE; i++ ) {
                        long l = o.bits[i] & ~bits[i] & mask[i];
                        if( l != 0L ) {
                            for( int j=0; j<64; j++ ) {
                                if( ( l & (1L<<j) ) != 0  ) {
                                    Node n = nodes[i*64+j];
                                    if( exclude.contains( n ) ) {
                                        l &= ~(1L<<j);
                                    }
                                }
                            }
                            if( l != 0L ) {
                                ret = true;
                                empty = false;
                                bits[i] |= l;
                            }
                        }
                    }
                }
                return ret;
            }
        }
        return super.addAll( other, exclude );
    }
    /** Calls v's visit method on all nodes in this set. */
    public final boolean forall( P2SetVisitor v ) {
        for( int i=0; i < SIZE; i++ ) {
            long bitsi = bits[i];
            if( bitsi != 0 ) for( int j=0; j<64; j++ ) {
                if( ( bitsi & (1L<<j) ) != 0  ) {
                    v.visit( nodes[i*64+j] );
                }
            }
        }
        boolean ret = v.getReturnValue();
        if( ret ) empty = false;
        return ret;
    }
    /** Adds n to this set, returns true if n was not already in this set. */
    public final boolean add( Node n ) {
        if( pag.getTypeManager().castNeverFails( n.getType(), type ) ) {
            return fastAdd( n );
        }
        return false;
    }
    /** Returns true iff the set contains n. */
    public final boolean contains( Node n ) {
        int id = -n.getId();
        return ( bits[ id/64 ] & 1L<<(id%64 ) ) != 0;
    }
    public static P2SetFactory getFactory() {
        return new P2SetFactory() {
            public PointsToSetInternal newSet( Type type, PAG pag ) {
                return new BitPointsToSet( type, pag );
            }
        };
    }

    public final static void delete() {
        nodes = null;
    }

    /* End of public methods. */
    /* End of package methods. */

    protected boolean fastAdd( Node n ) {
        int id = -n.getId();
        if( nodes[id] == null ) nodes[id] = n;
        if( ( bits[id/64] & 1L<<(id%64) ) != 0 ) return false;
        bits[id/64] |= 1L<<(id%64);
        empty = false;
        return true;
    }

    private static int SIZE = 0;
    private static Node[] nodes = null;
    private long[] bits = null;
    private boolean empty = true;
    private PAG pag = null;
}

