/* Soot - a J*va Optimization Framework
 * Copyright (C) 2002 Sable Research Group
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/*
 * Modified by the Sable Research Group and others 1997-1999.  
 * See the 'credits' file distributed with Soot for the complete list of
 * contributors.  (Soot is distributed at http://www.sable.mcgill.ca/soot)
 */

/* Reference Version: $SootVersion: 1.2.4.dev.18 $ */

/**
 * DotGraph provides an interface to SOOT for generating DOT language
 * for graphviz from ATT research lab.
 *
 * Intended ussage: virtualize CFG, graphes, etc...
 *
 * @author Feng Qian
 */

package soot.util;

import java.io.*;
import java.util.*;

public class DotGraph implements Renderable{
  
  /* allow a serialized drawing, following steps:
   * 1. new DotGraph
   * 2. draw(Directed)Edge / drawUndirectedEdge
   *    attachAttributes, addNode
   * 3. plot
   */
  private String  graphname;
  private boolean isSubGraph;

  private HashMap nodes;
  /* draw elements are sub graphs, edges, commands */
  private List    drawElements;
 
  private List    attributes;

  /**
   * Creates a new graph for drawing.
   * @param graphname, the file name with dot format will be generated
   */
  public DotGraph(String graphname) {
    this.graphname   = graphname;
    this.isSubGraph = false;
    this.nodes      = new HashMap(100);
    this.drawElements = new LinkedList();
    this.attributes   = new LinkedList();
  }
  
  /**
   * Generates the drawing on canvas to the dot file.
   */
  public void plot() {
    try {
      BufferedOutputStream out = 
	new BufferedOutputStream(new FileOutputStream(this.graphname+".dot"));
							  
      render(out, 0);
      out.close();

      System.out.println("Generate dot file in "+this.graphname+".dot");

    } catch (IOException ioe) {
    }
  }

  /**
   * Draws a directed edge
   * @param from, the source node
   * @param to, the end node
   * @return a graph edge
   */
  public DotGraphEdge drawEdge(String from, String to) {

    DotGraphNode src = (DotGraphNode)nodes.get(from);
    if (src == null) {
      src = new DotGraphNode(from);
      nodes.put(from, src);
    } 

    DotGraphNode dst = (DotGraphNode)nodes.get(to);
    if (dst == null) {
      dst = new DotGraphNode(to);
      nodes.put(to, dst);
    }

    DotGraphEdge edge = new DotGraphEdge(src, dst);
    
    this.drawElements.add(edge);
    
    return edge;
  }

  /**
   * Gets the graph node by name.
   * @param name, unique name of the node.
   */
  public DotGraphNode getNode(String name){
    DotGraphNode node = (DotGraphNode)nodes.get(name);
    if (node == null) {
      node = new DotGraphNode(name);
      nodes.put(name, node);
    }
    return node;
  }

  /**
   * Sets all node shapes, see the list of node shapes in DotGraphConstants.
   * @param shape, the node shape
   */
  public void setNodeShape(String shape){
    StringBuffer command = new StringBuffer("node [shape=");
    command.append(shape);
    command.append("];");
    this.drawElements.add(new DotGraphCommand(new String(command)));
  }

  /**
   * Sets all node styles
   * @param style, the node style
   */
  public void setNodeStyle(String style){
    StringBuffer command = new StringBuffer("node [style=");
    command.append(style);
    command.append("];");
    this.drawElements.add(new DotGraphCommand(new String(command)));
  }

  /**
   * sets the size of drawing area, in inches
   */
  public void setGraphSize(double width, double height){
    String size = "\""+width+","+height+"\"";
    this.setGraphAttribute("size", size);
  }

  /**
   * sets the pages size, once this is set, the generated graph
   * will be broken into several pages.
   */
  public void setPageSize(double width, double height){
    String size = "\""+width+", "+height+"\"";
    this.setGraphAttribute("page", size);
  }

  /**
   * sets the graph rotation angles
   */
  public void setOrientation(String orientation){
    this.setGraphAttribute("orientation", orientation);
  }

  /**
   * sets the graph label
   */
  public void setGraphLabel(String label){
    label = DotGraphUtility.replaceQuotes(label);
    label = DotGraphUtility.replaceReturns(label);
    this.setGraphAttribute("label", "\""+label+"\"");
  }

  /**
   * sets any general attributes
   */
  public void setGraphAttribute(String id, String value){
    this.attributes.add(new DotGraphAttribute(id, value));    
  }
  /**
   * draws an undirected edge
   * @param label1, label2
   */
  public void drawUndirectedEdge(String label1, String label2) {
  }

  /**
   * creates a sub graph.
   * @return the newly created sub graph.
   */
  public DotGraph createSubGraph(String label){
    // file name is used as label of sub graph.
    DotGraph subgraph = new DotGraph(label);
    subgraph.isSubGraph = true;

    this.drawElements.add(subgraph);

    return subgraph;
  }

  /* implements renderable interface. */
  public void render(OutputStream out, int indent) throws IOException{
    // header
    if (!isSubGraph) {
      DotGraphUtility.renderLine(out, "digraph \""+graphname+"\" {", indent);
    } else {
      DotGraphUtility.renderLine(out, "subgraph \""+graphname+"\" {", indent);
    }

    /* render graph attributes */
    Iterator attrIt = this.attributes.iterator();
    while (attrIt.hasNext()) {
      DotGraphAttribute attr = (DotGraphAttribute)attrIt.next();
      DotGraphUtility.renderLine(out, attr.toString()+";", indent+4);
    }

    /* render elements */
    Iterator elmntsIt = this.drawElements.iterator();
    while (elmntsIt.hasNext()) {
      Renderable element = (Renderable)elmntsIt.next();
      element.render(out, indent+4);
    }

    // close the description
    DotGraphUtility.renderLine(out, "}", indent);
  }
}
